==> nfkit-v1.0.26.zip: nfkit-v1.0.26/src/app-context/app-context.ts <==
import { AnyClass, Empty } from '../types';
import {
  AppContext,
  AppServiceClass,
  AppProvideArgs,
  AppProvidedMerged,
  AppProvideOptions,
  AppContextUsed,
  AppContextState,
} from './types';

const ProvidePrefix = 'provide:';

const getMethodDescriptor = (cls: AnyClass, key: PropertyKey) => {
  let proto = cls.prototype;
  while (proto && proto !== Object.prototype) {
    const desc = Object.getOwnPropertyDescriptor(proto, key as any);
    if (desc) return desc;
    proto = Object.getPrototypeOf(proto);
  }
  return undefined;
};

type ProvideRecord = {
  classRef: AnyClass;
  factory: () => any;
  provideKey?: string;
};

type LoadEntry = {
  classRef: AnyClass;
  inst: any;
};

type ObjectStep = (ctx: AppContextCore<any, any>) => void;

export class AppContextCore<Cur = Empty, Req = Empty> {
  private __current: Cur;
  private __required: Req;

  private provideRecords: ProvideRecord[] = [];
  private registry = new Map<string | AnyClass, LoadEntry>();
  private loadSeq: LoadEntry[] = [];
  private objectSteps: ObjectStep[] = [];
  private started = false;

  provide<
    C extends AppServiceClass<Cur, Req>,
    const P extends string = '',
    const M extends (keyof InstanceType<C>)[] = [],
  >(
    cls: C,
    ...args: AppProvideArgs<Cur, Req, C, P, M>
  ): AppProvidedMerged<Cur, Req, C, P, M> {
    const last = args[args.length - 1] as any;
    const hasOptions =
      !!last &&
      typeof last === 'object' &&
      ('provide' in last ||
        'merge' in last ||
        'useValue' in last ||
        'useFactory' in last ||
        'useClass' in last);

    const options = (hasOptions ? last : undefined) as
      | AppProvideOptions<Cur, Req, C, P, M>
      | undefined;

    const _args = (
      hasOptions ? args.slice(0, -1) : args
    ) as ConstructorParameters<C>;

    const classRef = cls as unknown as AnyClass;
    const provideKey = options?.provide
      ? ProvidePrefix + String(options.provide)
      : undefined;

    // Create factory function that will be called during start()
    const factory = () =>
      options?.useValue ??
      (options?.useFactory
        ? options.useFactory(this as any, ..._args)
        : new (options?.useClass ?? cls)(this as any, ..._args));

    // Record the provide configuration
    this.provideRecords.push({
      classRef,
      factory,
      provideKey,
    });

    // Set up property accessors if needed
    if (options?.provide) {
      const prop = options.provide;
      const step: ObjectStep = (ctx) => {
        Object.defineProperty(ctx, prop, {
          enumerable: true,
          configurable: true,
          get: () => {
            const currentEntry = ctx.registry.get(classRef);
            return currentEntry?.inst;
          },
        });
      };
      step(this);
      this.objectSteps.push(step);
    }

    if (options?.merge?.length) {
      for (const key of options.merge) {
        const desc = getMethodDescriptor(cls, key);
        const isMethod = !!desc && typeof desc.value === 'function';

        const step: ObjectStep = (ctx) => {
          if (isMethod) {
            Object.defineProperty(ctx, key, {
              enumerable: true,
              configurable: true,
              get: () => {
                const currentEntry = ctx.registry.get(classRef);
                const currentInst = currentEntry?.inst;
                const fn = (currentInst as any)?.[key];
                return typeof fn === 'function' ? fn.bind(currentInst) : fn;
              },
            });
            return;
          }

          Object.defineProperty(ctx, key, {
            enumerable: true,
            configurable: true,
            get: () => {
              const currentEntry = ctx.registry.get(classRef);
              return (currentEntry?.inst as any)?.[key];
            },
            set: (value: any) => {
              const currentEntry = ctx.registry.get(classRef);
              if (!currentEntry) return;
              (currentEntry.inst as any)[key] = value;
            },
          });
        };
        step(this);
        this.objectSteps.push(step);
      }
    }

    return this as any;
  }

  get<R>(
    cls:
      | AppServiceClass<Cur, Req, any, R>
      | (() => AppServiceClass<Cur, Req, any, R>),
  ): R {
    let key = cls as unknown as AnyClass;
    if (
      !this.registry.has(key) &&
      typeof cls === 'function' &&
      !(cls as any).prototype
    ) {
      key = (cls as () => AppServiceClass<Cur, Req, any, R>)() as AnyClass;
    }

    if (!this.registry.has(key)) {
      throw new Error(`Service not provided: ${key?.name ?? cls.name}`);
    }

    const entry = this.registry.get(key)!;
    return entry.inst as any;
  }

  /**
   * @deprecated Use get() instead. getAsync() is no longer needed as all services are loaded synchronously during start().
   */
  async getAsync<R>(
    cls:
      | AppServiceClass<Cur, Req, any, R>
      | (() => AppServiceClass<Cur, Req, any, R>),
  ): Promise<R> {
    return this.get(cls);
  }

  use<const Ctxes extends AppContext<any, any>[]>(
    ...ctxes: Ctxes
  ): AppContextUsed<Cur, Req, Ctxes> {
    for (const ctx of ctxes) {
      const other = ctx as any as AppContextCore<any, any>;

      // Copy provide records
      if (Array.isArray(other?.provideRecords)) {
        this.provideRecords.push(...other.provideRecords);
      }

      // Copy and apply object steps
      if (Array.isArray(other?.objectSteps)) {
        this.objectSteps.push(...other.objectSteps);
        for (const step of other.objectSteps) {
          step(this);
        }
      }

      // If the other context has already started, copy its loaded entries
      if (other?.started) {
        if (Array.isArray(other?.loadSeq)) {
          this.loadSeq.push(...other.loadSeq);
        }
        if (other?.registry instanceof Map) {
          for (const [key, value] of other.registry.entries()) {
            this.registry.set(key, value);
          }
        }
      }
    }

    return this as any;
  }

  define(): AppContext<Cur, Req> {
    return this as any;
  }

  async start(): Promise<Empty extends Req ? AppContext<Cur, Req> : never> {
    if (this.started) {
      return this as any;
    }

    // Create all instances
    for (const record of this.provideRecords) {
      const inst = record.factory();
      const entry: LoadEntry = {
        classRef: record.classRef,
        inst,
      };
      this.registry.set(record.classRef, entry);
      if (record.provideKey) {
        this.registry.set(record.provideKey, entry);
      }
      this.loadSeq.push(entry);
    }

    // Resolve all promises
    for (const entry of this.loadSeq) {
      if (entry.inst && typeof entry.inst.then === 'function') {
        entry.inst = await entry.inst;
      }
    }

    // Call init on all instances
    for (const entry of this.loadSeq) {
      const inst = entry.inst;
      if (inst && typeof inst.init === 'function') {
        await inst.init();
      }
    }

    this.started = true;
    return this as any;
  }
}

export const createAppContext = <Req = Empty>() =>
  new AppContextCore<Empty, Req>() as AppContext<Empty, Req>;

==> nfkit-v1.0.28.zip: nfkit-v1.0.28/src/app-context/app-context.ts <==
import { AnyClass, Empty } from '../types';
import {
  AppContext,
  AppServiceClass,
  AppProvideArgs,
  AppProvidedMerged,
  AppProvideOptions,
  AppContextUsed,
  AppContextState,
} from './types';

const ProvidePrefix = 'provide:';

const getMethodDescriptor = (cls: AnyClass, key: PropertyKey) => {
  let proto = cls.prototype;
  while (proto && proto !== Object.prototype) {
    const desc = Object.getOwnPropertyDescriptor(proto, key as any);
    if (desc) return desc;
    proto = Object.getPrototypeOf(proto);
  }
  return undefined;
};

type ProvideRecord = {
  classRef: AnyClass;
  factory: (ctx: AppContextCore<any, any>) => any;
};

type LoadEntry = {
  classRef: AnyClass;
  inst: any;
};

type ObjectStep = (ctx: AppContextCore<any, any>) => void;

export class AppContextCore<Cur = Empty, Req = Empty> {
  private __current: Cur;
  private __required: Req;

  private provideRecords: ProvideRecord[] = [];
  private registry = new Map<string | AnyClass, LoadEntry>();
  private objectSteps: ObjectStep[] = [];
  started = false;

  provide<
    C extends AppServiceClass<Cur, Req>,
    const P extends string = '',
    const M extends (keyof InstanceType<C>)[] = [],
  >(
    cls: C,
    ...args: AppProvideArgs<Cur, Req, C, P, M>
  ): AppProvidedMerged<Cur, Req, C, P, M> {
    const last = args[args.length - 1] as any;
    const hasOptions =
      !!last &&
      typeof last === 'object' &&
      ('provide' in last ||
        'merge' in last ||
        'useValue' in last ||
        'useFactory' in last ||
        'useClass' in last);

    const options = (hasOptions ? last : undefined) as
      | AppProvideOptions<Cur, Req, C, P, M>
      | undefined;

    const _args = (
      hasOptions ? args.slice(0, -1) : args
    ) as ConstructorParameters<C>;

    const classRef = cls as unknown as AnyClass;

    // Create factory function that will be called during start() with the target ctx.
    const factory = (ctx: AppContextCore<any, any>) =>
      options?.useValue ??
      (options?.useFactory
        ? options.useFactory(ctx as any, ..._args)
        : new (options?.useClass ?? cls)(ctx as any, ..._args));

    // Record the provide configuration
    this.provideRecords.push({
      classRef,
      factory,
    });

    // Set up property accessors if needed
    if (options?.provide) {
      const prop = options.provide;
      const step: ObjectStep = (ctx) => {
        Object.defineProperty(ctx, prop, {
          enumerable: true,
          configurable: true,
          get: () => {
            const currentEntry = ctx.registry.get(classRef);
            return currentEntry?.inst;
          },
        });
      };
      step(this);
      this.objectSteps.push(step);
    }

    if (options?.merge?.length) {
      for (const key of options.merge) {
        const desc = getMethodDescriptor(cls, key);
        const isMethod = !!desc && typeof desc.value === 'function';

        const step: ObjectStep = (ctx) => {
          if (isMethod) {
            Object.defineProperty(ctx, key, {
              enumerable: true,
              configurable: true,
              get: () => {
                const currentEntry = ctx.registry.get(classRef);
                const currentInst = currentEntry?.inst;
                const fn = (currentInst as any)?.[key];
                return typeof fn === 'function' ? fn.bind(currentInst) : fn;
              },
            });
            return;
          }

          Object.defineProperty(ctx, key, {
            enumerable: true,
            configurable: true,
            get: () => {
              const currentEntry = ctx.registry.get(classRef);
              return (currentEntry?.inst as any)?.[key];
            },
            set: (value: any) => {
              const currentEntry = ctx.registry.get(classRef);
              if (!currentEntry) return;
              (currentEntry.inst as any)[key] = value;
            },
          });
        };
        step(this);
        this.objectSteps.push(step);
      }
    }

    return this as any;
  }

  get<R>(
    cls:
      | AppServiceClass<Cur, Req, any, R>
      | (() => AppServiceClass<Cur, Req, any, R>),
  ): R {
    let key = cls as unknown as AnyClass;
    if (
      !this.registry.has(key) &&
      typeof cls === 'function' &&
      !(cls as any).prototype
    ) {
      key = (cls as () => AppServiceClass<Cur, Req, any, R>)() as AnyClass;
    }

    if (!this.registry.has(key)) {
      throw new Error(`Service not provided: ${key?.name ?? cls.name}`);
    }

    const entry = this.registry.get(key)!;
    return entry.inst as any;
  }

  /**
   * @deprecated Use get() instead. getAsync() is no longer needed as all services are loaded synchronously during start().
   */
  async getAsync<R>(
    cls:
      | AppServiceClass<Cur, Req, any, R>
      | (() => AppServiceClass<Cur, Req, any, R>),
  ): Promise<R> {
    return this.get(cls);
  }

  use<const Ctxes extends AppContext<any, any>[]>(
    ...ctxes: Ctxes
  ): AppContextUsed<Cur, Req, Ctxes> {
    for (const ctx of ctxes) {
      const other = ctx as any as AppContextCore<any, any>;

      if (this.started && !other?.started) {
        throw new Error(
          'Cannot use an unstarted context into a started context.',
        );
      }

      // Copy provide records
      if (Array.isArray(other?.provideRecords)) {
        this.provideRecords.push(...other.provideRecords);
      }

      // Copy and apply object steps
      if (Array.isArray(other?.objectSteps)) {
        this.objectSteps.push(...other.objectSteps);
        for (const step of other.objectSteps) {
          step(this);
        }
      }

      // If the other context has already started, copy loaded entries only.
      // They should remain initialized by the source context and must not be re-initialized.
      if (other?.started) {
        if (other?.registry instanceof Map) {
          for (const [key, value] of other.registry.entries()) {
            this.registry.set(key, value);
          }
        }
      }
    }

    return this as any;
  }

  define(): AppContext<Cur, Req> {
    return this as any;
  }

  async start(): Promise<Empty extends Req ? AppContext<Cur, Req> : never> {
    if (this.started) {
      return this as any;
    }

    const startedEntries: LoadEntry[] = [];
    const preloadedKeys = new Set(this.registry.keys());

    // Create all instances
    for (const record of this.provideRecords) {
      if (preloadedKeys.has(record.classRef)) {
        continue;
      }

      const inst = record.factory(this);
      const entry: LoadEntry = {
        classRef: record.classRef,
        inst,
      };
      this.registry.set(record.classRef, entry);
      startedEntries.push(entry);
    }

    // Resolve all promises created in this start().
    for (const entry of startedEntries) {
      if (entry.inst && typeof entry.inst.then === 'function') {
        entry.inst = await entry.inst;
      }
    }

    // Init only instances created in this start().
    for (const entry of startedEntries) {
      const inst = entry.inst;
      if (inst && typeof inst.init === 'function') {
        await inst.init();
      }
    }

    this.started = true;
    return this as any;
  }
}

export const createAppContext = <Req = Empty>() =>
  new AppContextCore<Empty, Req>() as AppContext<Empty, Req>;
